"""MiniCPM-o 顶层工具包 `minicpmo`。

这个包聚合了语音 / 音频相关子包，并预留统一的 utils 入口：

- 顶层子包：
  - cosyvoice
  - stepaudio2
  - matcha
  - s3tokenizer
- 工具入口：
  - from minicpmo.utils import ...
"""

from .version import __version__
